/*******************************************************************************
DeskCalc. Copyright 1997-2000 Wunderbear Software. See below for license info.
$Id: CalcView.cpp,v 1.10 1998/12/03 08:29:39 timmy Exp $
Code Classification: C++, BeOS

This component implements actual functional GUI interface of the calculator.

Copyright (c) 1997-2000 Wunderbear Software. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1) Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

2) Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.

3) Neither the name of the Wunderbear Software nor the names of its contributors
may be used to endorse or promote products derived from this software without
specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COMPANY OR CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*******************************************************************************/


#include "CalcView.h"


// Standard Includes

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>


// BeOS Includes

#ifndef _MESSAGE_H
	#include <app/Message.h>
#endif

#ifndef _CLIPBOARD_H
	#include <app/Clipboard.h>
#endif

#ifndef _FONT_H
	#include <interface/Font.h>
#endif

#ifndef _POINT_H
	#include <interface/Point.h>
#endif

#ifndef _RECT_H
	#include <interface/Rect.h>
#endif

#ifndef _POP_UP_MENU_H
	#include <interface/PopUpMenu.h>
#endif

#ifndef _PLAY_SOUND_H
	#include <media/PlaySound.h>
#endif

#ifndef _BEEP_H
	#include <support/Beep.h>
#endif

#ifndef __BSTRING__
	#include <support/String.h>
#endif


// Library Includes

#ifndef PARSER_HEADER
	#include "Parser.h"
#endif


// RView Includes

/*
#ifndef _R_ABOUT_BOX_H
	#include <RAboutBox.h>
#endif
*/


// Application Includes

#ifndef _CALC_OPTIONS_WINDOW_H
	#include "CalcOptionsWindow.h"
#endif


// Internal Constants

const uint8 K_COLOR_OFFSET = 32;
const float K_FONT_YPROP = 0.6f;
const float K_DISPLAY_YPROP = 0.2;


// Code!

/*******************************************************************************
<func>
Class:		CalcView
Function:	static CalcView *Instantiate(BMessage *archive);
Summary:	Constructor.
Params:		archive		data to instantiate from
Returns:	Instantiated CalcView.
Author:		timmy
Created:	1997.10.14
History:
</func>
*******************************************************************************/
CalcView *CalcView::Instantiate(BMessage *archive){ 

	if (!validate_instantiation(archive, "CalcView"))
		return NULL; 
	return new CalcView(archive);
	
} // end CalcView::Instantiate()


/*******************************************************************************
<func>
Class:		CalcView
Function:	CalcView(BRect frame,
				const char *calcDesc,
				rgb_color rgbBaseColor);
Summary:	Constructor.
Params:		frame		on-screen window bounds
			calcDesc	calculator description string
Author:		timmy
Created:	1997.10.02
History:
1997.10.18 timmy Added pop-up menu initialisation.
1997.10.20 timmy Removed B_FRAME_EVENTS flag from BView ctor.
1997.10.28 timmy Added initialisation and setup of mOptions.
1998.11.07 timmy Fixed mCalcDesc allocation size bug.
1999.06.10 timmy Altered view name string.
</func>
*******************************************************************************/
CalcView::CalcView(BRect frame,
		const char *calcDesc,
		rgb_color rgbBaseColor)
	:	BView(frame,
			"calcView",
			B_FOLLOW_ALL_SIDES,
			B_WILL_DRAW/*|B_FRAME_EVENTS*/|B_FULL_UPDATE_ON_RESIZE),
		mCols(5),
		mRows(4),
		mRGBBaseColor(rgbBaseColor),
		mWidth(frame.right - frame.left),
		mHeight(frame.bottom - frame.top),
		mCalcDesc(NULL),
		mKeypad(NULL),
		mDisplayText(NULL),
		mAboutItem(NULL),
		mOptionsItem(NULL),
		mPopUpMenu(NULL),
		mOptions(NULL),
		mOptionsWindow(NULL)
		{
	
	// save calculator description for reference
	mCalcDesc = strdup(calcDesc);
	
	// parse calculator description
	ParseCalcDesc(calcDesc);
	
	// create options with default settings
	mOptions = new CalcOptions;
	
	// Colorize based on base color.
	Colorize();
	
	// tell the app server not to erase our b/g
	SetViewColor(B_TRANSPARENT_32_BIT);
	
	// create (empty) string for display text
	mDisplayText = new BString();
	
	// create pop-up menu system
	CreatePopUpMenu();
	
} // end CalcView::CalcView()


/*******************************************************************************
<func>
Class:		CalcView
Function:	CalcView(BMessage *archive);
Summary:	Construct from archive.
Params:		archive		Archived CalcView in message.
Author:		timmy
Created:	1997.10.14
History:
1998.11.07 timmy Fixed mCalcDesc allocation size bug.
</func>
*******************************************************************************/
CalcView::CalcView(BMessage *archive)
	:	BView(archive),
		mCalcDesc(NULL),
		mDisplayText(NULL),
		mAboutItem(NULL),
		mOptionsItem(NULL),
		mPopUpMenu(NULL),
		mOptionsWindow(NULL)
		{
	
	// read data from archive

	// record calculator description
	const char *calcDesc;
	if (archive->FindString("calcDesc", &calcDesc) < B_OK){
		puts("Missing calcDesc from CalcView archive!\n");
		exit(EXIT_FAILURE);
		} // end if

	// save calculator description for reference
	mCalcDesc = strdup(calcDesc);

	// read grid dimensions
	if (archive->FindInt16("cols", &mCols) < B_OK) mCols = 5;
	if (archive->FindInt16("rows", &mRows) < B_OK) mRows = 4;
	
	// read pixel dimensions
	if (archive->FindFloat("width", &mWidth) < B_OK) mWidth = 256.0;
	if (archive->FindFloat("height", &mHeight) < B_OK) mHeight = 160.0;
	
	// read color scheme
	const void *rgbBaseColor;
	ssize_t size;
	if (archive->FindData("rgbBaseColor", B_RGB_COLOR_TYPE,
		&rgbBaseColor, &size) < B_OK){
			rgb_color calcBaseColor = {128, 128, 128, 0};
			mRGBBaseColor = calcBaseColor;
			puts("Missing rgbBaseColor from CalcView archive!\n");
			} // end if
	else {
		mRGBBaseColor = *((rgb_color *) rgbBaseColor);
		} // end else
		
	/* NOT YET USED
	if (archive->FindData("rgbDisplay", B_RGB_COLOR_TYPE,
		&mRGBDisplay, sizeof(rgb_color)) < B_OK) return B_ERROR;
	*/
	
	// load options
	const void *options;
	mOptions = new CalcOptions;
	if (archive->FindData("options", B_RAW_TYPE,
		&options, &size) < B_OK){
			puts("Missing options from CalcView archive.\n");
			} // end if
	else {
		memcpy(mOptions, options, size);
		} // end else
	
	// load display text
	const char *display;
	if (archive->FindString("displayText", &display) < B_OK){
		puts("Missing display text from CalcView archive.\n");
		// create (empty) string for display text
		mDisplayText = new BString();
		} // end if
	else {
		// create string for display text
		mDisplayText = new BString(display);
		} // end else
	
	// load frame dimensions
	BRect frame = Frame();
	mWidth = frame.right - frame.left;
	mHeight = frame.bottom - frame.top;
	
	// parse calculator description
	ParseCalcDesc(calcDesc);
	
	// Colorize based on base color.
	Colorize();
	
	// tell the app server not to erase our b/g
	SetViewColor(B_TRANSPARENT_32_BIT);
	
	// create pop-up menu system
	CreatePopUpMenu();
	
} // end CalcView::CalcView()

 
/*******************************************************************************
<func>
Class:		CalcView
Function:	virtual ~CalcView();
Summary:	Destructor.
Author:		timmy
Created:	1997.10.02
History:
1997.10.04 timmy Added deallocation for mKeypad.
1997.10.05 timmy Added deallocation for mDisplayText.
1997.10.28 timmy Now deallocates mOptions.
1997.10.28 timmy Added deallocation for mCalcDesc.
</func>
*******************************************************************************/
CalcView::~CalcView(){

	delete mKeypad;
	delete mDisplayText;
	delete mOptions;
	free(mCalcDesc);

} // end CalcView::~CalcView()


/*******************************************************************************
<func>
Class:		CalcView
Function:	virtual void MessageReceived(BMessage *message);
Summary:	Message received hook.
Params:		message		the incoming message
Author:		timmy
Created:	1997.10.12
History:
1997.10.17 timmy Added remoteness test for drag & drop.
1998.07.26 timmy Updated for R3 mwcc compiler.
1998.11.07 timmy Fixed switch order bug.
</func>
*******************************************************************************/
void CalcView::MessageReceived(BMessage *message){
	
	// debug
	//message->PrintToStream();
	
	// check if message was dropped
	if (message->WasDropped()){
	
		// pass message on to paste
		if (message->IsSourceRemote()) Paste(message);
		
		} // end if
	
	else {

		// act on posted message type
		switch (message->what){
		
			// handle "cut"
			case B_CUT:
			Cut();
			break;
			
			// handle copy
			case B_COPY:
			Copy();
			break;
			
			// handle paste
			case B_PASTE:
			// access system clipboard
			if (be_clipboard->Lock()){ 
				BMessage *clipper = be_clipboard->Data();
				//clipper->PrintToStream();
				Paste(clipper);
				be_clipboard->Unlock(); 
				} // end if	
			break;
			
			// (replicant) about box requested
			case B_ABOUT_REQUESTED:
			AboutRequested();
			break;
			
			// calculator options window requested
			case K_OPTIONS_REQUESTED:
			{	BRect frame(30.0f, 50.0f, 230.0, 200.0);
				mOptionsWindow = new CalcOptionsWindow(frame, mOptions);
				mOptionsWindow->Show();
				} // end block
			break;
			
			default:
			BView::MessageReceived(message);
			
			} // end switch

		} // end else

} // end CalcView::MessageReceived()


/*******************************************************************************
<func>
Class:		CalcView
Function:	void Draw(BRect updateRect);
Summary:	Render view hook.
Params:		updateRect	area which requires update
Author:		timmy
Created:	1997.10.02
History:
1997.10.05 timmy Added render for display text.
1997.10.18 timmy Fixed line array render bug.
</func>
*******************************************************************************/
void CalcView::Draw(BRect updateRect){

	rgb_color rgbWhite = {255, 255, 255, 0};

	BRect frame = Frame();
	mWidth = frame.right - frame.left;
	mHeight = frame.bottom - frame.top;

	// calculate grid sizes
	float sizeDisp = mHeight * K_DISPLAY_YPROP;
	float sizeCol = mWidth/(float)mCols;
	float sizeRow = (mHeight - sizeDisp)/(float)mRows;
	
	// setup areas
	BRect displayRect(0.0, 0.0, mWidth, sizeDisp);
	BRect keypadRect(0.0, sizeDisp, mWidth, mHeight);
	
	SetFont(be_bold_font);
	
	// ****** DISPLAY Area
	if (updateRect.Intersects(displayRect)){
	
		// paint display b/g
		SetHighColor(mRGBDisplay);
		FillRect(updateRect & displayRect);
		
		// render display text
		SetHighColor(rgbWhite);
		SetLowColor(mRGBDisplay);
		SetDrawingMode(B_OP_COPY);
		SetFontSize(sizeDisp * K_FONT_YPROP);
		float baselineOffset = sizeDisp * (1.0 - K_FONT_YPROP) * 0.5;
		DrawString(mDisplayText->String(),
					BPoint(mWidth - StringWidth(mDisplayText->String()),
						sizeDisp - baselineOffset));
		
		} // end if
		
	
	// ****** KEYPAD Area
	if (updateRect.Intersects(keypadRect)){
	
		// paint keypad b/g
		SetHighColor(mRGBBaseColor);
		FillRect(updateRect & keypadRect);
		
		// render key main grid
		BeginLineArray((mCols + mRows) << 1 + 1);
		
		// render cols
		AddLine(BPoint(0.0, sizeDisp),
			BPoint(0.0, mHeight),
			mRGBLightEdge);
		for (int col = 1; col < mCols; col++){
			AddLine(BPoint(col * sizeCol - 1.0, sizeDisp),
				BPoint(col * sizeCol - 1.0, mHeight),
				mRGBDarkEdge);
			AddLine(BPoint(col * sizeCol, sizeDisp),
				BPoint(col * sizeCol, mHeight),
				mRGBLightEdge);
			} // end for
		AddLine(BPoint(mCols * sizeCol, sizeDisp),
			BPoint(mCols * sizeCol, mHeight),
			mRGBDarkEdge);
			
		// render rows
		for (int row = 0; row < mRows; row++){
			AddLine(BPoint(0.0, sizeDisp + row * sizeRow - 1.0),
				BPoint(mWidth, sizeDisp + row * sizeRow - 1.0),
				mRGBDarkEdge);
			AddLine(BPoint(0.0, sizeDisp + row * sizeRow),
				BPoint(mWidth, sizeDisp + row * sizeRow),
				mRGBLightEdge);
			} // end for
		AddLine(BPoint(0.0, sizeDisp + mRows * sizeRow),
			BPoint(mWidth, sizeDisp + mRows * sizeRow),
			mRGBDarkEdge);
			
		// main grid complete
		EndLineArray();
		
		// render key symbols
		float halfSizeCol = sizeCol * 0.5f;
		SetHighColor(rgbWhite);
		SetLowColor(mRGBBaseColor);
		SetDrawingMode(B_OP_COPY);
		SetFontSize(((mHeight - sizeDisp)/(float)mRows) * K_FONT_YPROP);
		float baselineOffset = ((mHeight - sizeDisp)/(float)mRows)
			* (1.0 - K_FONT_YPROP) * 0.5;
		CalcKey *key = mKeypad;
		for (int row = 0; row < mRows; row++){
			for (int col = 0; col < mCols; col++){
				float halfSymbolWidth = StringWidth(key->label) * 0.5f;
				DrawString(key->label,
					BPoint(col * sizeCol + halfSizeCol - halfSymbolWidth,
						sizeDisp + (row + 1) * sizeRow - baselineOffset));
				key++;
				} // end for
			} // end for

		} // end if
	
} // end CalcView::Draw()


/*******************************************************************************
<func>
Class:		CalcView
Function:	virtual void MouseDown(BPoint point);
Summary:	Mouse button pressed hook.
Params:		point	location of mouse down
Author:		timmy
Created:	1997.10.05
History:
1997.10.16 timmy Fixed display click bug.
1997.10.18 timmy Added pop-up menu access.
</func>
*******************************************************************************/
void CalcView::MouseDown(BPoint point){

	// ensure this view is the current focus
	MakeFocus();

	// read mouse buttons state
	int32 buttons = 0;
	Window()->CurrentMessage()->FindInt32("buttons", &buttons);

	// check for non-primary mouse button
	if ((B_PRIMARY_MOUSE_BUTTON & buttons) == 0){
		BMenuItem *selected;
		if ((selected = mPopUpMenu->Go(ConvertToScreen(point))) != NULL){
			MessageReceived(selected->Message());
			} // end if
		return;
		} // end if

	// calculate grid sizes
	float sizeDisp = mHeight * K_DISPLAY_YPROP;
	float sizeCol = mWidth/(float)mCols;
	float sizeRow = (mHeight - sizeDisp)/(float)mRows;
	
	// click on display
	if ((point.y - sizeDisp) < 0.0){
		
		// only drag if there's some text!
		if (mDisplayText->Length() > 0){
		
			// assemble drag message
			BMessage dragmsg(B_MIME_DATA);
			dragmsg.AddData("text/plain",
				B_MIME_TYPE,
				mDisplayText->String(), 
				mDisplayText->Length());
				
			// initiate drag & drop
			SetFontSize(sizeDisp * K_FONT_YPROP);
			float left = mWidth;
			float textWidth = StringWidth(mDisplayText->String());
			if (textWidth < mWidth) left -= textWidth;
			else left = 0;
			BRect displayRect(left, 0.0, mWidth, sizeDisp);
			DragMessage(&dragmsg, displayRect);
			
			} // end if
		
		} // end if
	
	// click on keypad
	else {

		// calculate location within grid
		int gridCol = (int)(point.x/sizeCol);
		int gridRow = (int)((point.y - sizeDisp)/sizeRow);
		
		// check limits
		if ((gridCol >= 0) && (gridCol < mCols) &
			(gridRow >= 0) && (gridRow < mRows)){
		
			// process key press
			PressKey(gridRow * mCols + gridCol);
			
			} // end if

		} // end else

} // end CalcView::MouseDown()


/*******************************************************************************
<func>
Class:		CalcView
Function:	virtual void KeyDown(const char *bytes, int32 numBytes);
Summary:	Key depressed hook.
Params:		bytes		key press data (unicode UTF-8)
			numBytes	size of key data
Author:		timmy
Created:	1997.10.05
History:
1997.10.19 timmy Added bracket translation.
</func>
*******************************************************************************/
void CalcView::KeyDown(const char *bytes, int32 numBytes){
 
 	// if single byte character...
	if (numBytes == 1){ 
		
		//printf("Key pressed: %c\n", bytes[0]);
		
		switch (bytes[0]){
		
			case B_ENTER:
			// translate to evaluate key
			PressKey("=");
			break;
			
			case B_LEFT_ARROW:
			case B_BACKSPACE:
			// translate to backspace key
			PressKey("BS");
			break;
			
			case B_SPACE:
			case B_ESCAPE:
			case 'c':	// hack!
			// translate to clear key
			PressKey("C");
			break;
			
			// bracket translation
			case '[':
			case '{':
			PressKey("(");
			break;
			case ']':
			case '}':
			PressKey(")");
			break;
			
			default:
			{ // block
				// scan the keymap array for match
				int keys = mRows * mCols;
				for (int i=0; i<keys; i++){
					if (mKeypad[i].keymap[0] == bytes[0]){
						PressKey(i);
						return;
						} // end if
					} // end for
				} // end block
			break;
			
			} // end switch
		
		} // end if

} // end CalcView::KeyDown()


/*******************************************************************************
<func>
Class:		CalcView
Function:	virtual void MakeFocus(bool focused = true);
Summary:	View receives focus hook.
Params:		focused		is this view now in focus?
Author:		timmy
Created:	1997.10.25
History:
1997.10.28 timmy Made auto num-lock conditional on mOptions.
1998.07.26 timmy Updated for R3 mwcc compiler.
</func>
*******************************************************************************/
void CalcView::MakeFocus(bool focused){

	if (focused){
		// set num lock
		if (mOptions->mAutoNumLock){
			set_keyboard_locks(B_NUM_LOCK |
				(modifiers() & (B_CAPS_LOCK | B_SCROLL_LOCK)));
			} // end if		
		} // end if

	// pass on request to parent
	BView::MakeFocus(focused);

} // end CalcView::MakeFocus()


/*******************************************************************************
<func>
Class:		CalcView
Function:	virtual void AboutRequested();
Summary:	Present about box for view (replicant).
Author:		timmy
Created:	1997.10.18
History:	-
</func>
*******************************************************************************/
void CalcView::AboutRequested(){

/*
	BRect frame(100.0, 100.0, 300.0, 280.0);
	new RAboutBox(frame, "DeskCalc v1.3.0", "code: timmy\nqa: rosalie\n\n"
		B_UTF8_COPYRIGHT"1997 R3 Software Ltd.\nAll Rights Reserved.",
		RAboutBox::R_SHOW_ON_OPEN);
*/

} // end CalcView::AboutRequested()


/*******************************************************************************
<func>
Class:		CalcView
Function:	virtual status_t Archive(BMessage *archive, bool deep = true) const;
Summary:	Archive this view.
Params:		archive		message to store object into
			deep		archive child objects?
Returns:	Status of archive operation.
Author:		timmy
Created:	1997.10.13
History:
1998.07.28 timmy Updated for R3 mwcc compiler.
</func>
*******************************************************************************/
status_t CalcView::Archive(BMessage *archive, bool deep) const {

	// passed on request to parent
	status_t ret;
	if ((ret = BView::Archive(archive, deep)) < B_OK) return ret;

	// add class and replicant info
	if (archive->AddString("class", "CalcView") < B_OK) return B_ERROR;
	// TODO: bad hardcode here!
    if (archive->AddString("add_on", "application/x-vnd.r3-deskcalc") < B_OK)
    	return B_ERROR;

	// record grid dimensions
	if (archive->AddInt16("cols", mCols) < B_OK) return B_ERROR;
	if (archive->AddInt16("rows", mRows) < B_OK) return B_ERROR;
	
	// record pixel dimensions
	if (archive->AddFloat("width", mWidth) < B_OK) return B_ERROR;
	if (archive->AddFloat("height", mHeight) < B_OK) return B_ERROR;
	
	// record color scheme
	if (archive->AddData("rgbBaseColor", B_RGB_COLOR_TYPE,
		&mRGBBaseColor, sizeof(rgb_color)) < B_OK) return B_ERROR;
	if (archive->AddData("rgbDisplay", B_RGB_COLOR_TYPE,
		&mRGBDisplay, sizeof(rgb_color)) < B_OK) return B_ERROR;

	// record current options
	if (archive->AddData("options", B_RAW_TYPE,
		mOptions, sizeof(CalcOptions)) < B_OK) return B_ERROR;

	// record display text
	if (archive->AddString("displayText", mDisplayText->String()) < B_OK)
		return B_ERROR;

	// record calculator description
	if (archive->AddString("calcDesc", mCalcDesc) < B_OK) return B_ERROR;

	// all was well
	return B_OK;

} // end CalcView::Archive()


/*******************************************************************************
<func>
Class:		CalcView
Function:	void Cut();
Summary:	Cut contents of view to system clipboard.
Author:		timmy
Created:	1997.10.16
History:	-
</func>
*******************************************************************************/
void CalcView::Cut(){

	Copy();	// copy data to clipboard
	mDisplayText->SetTo(""); // remove data

	// redraw display
	float sizeDisp = mHeight * K_DISPLAY_YPROP;
	BRect displayRect(0.0, 0.0, mWidth, sizeDisp);
	Draw(displayRect);

} // end CalcView::Cut()


/*******************************************************************************
<func>
Class:		CalcView
Function:	void Copy();
Summary:	Copy contents of view to system clipboard.
Author:		timmy
Created:	1997.10.16
History:	-
</func>
*******************************************************************************/
void CalcView::Copy(){

	// access system clipboard
	if (be_clipboard->Lock()){ 
		be_clipboard->Clear(); 
		BMessage *clipper = be_clipboard->Data(); 
		clipper->what = B_MIME_DATA;
		//clipper->AddString("text/plain", mDisplayText->String());
		// TODO: should check return for errors!
		clipper->AddData("text/plain",
			B_MIME_TYPE,
			mDisplayText->String(), 
			mDisplayText->Length());
		//clipper->PrintToStream();
		be_clipboard->Commit(); 
		be_clipboard->Unlock(); 
		} // end if
	
	else beep();

} // end CalcView::Copy()

 
/*******************************************************************************
<func>
Class:		CalcView
Function:	void Paste();
Summary:	Paste contents of system clipboard to view.
Params:		message		incoming data
Author:		timmy
Created:	1997.10.16
History:	-
</func>
*******************************************************************************/
void CalcView::Paste(BMessage *message){
	
	// handle color drops first		
	// read incoming color
	const void *dropColor;
	ssize_t dataSize;
	if (message->FindData("RGBColor", B_RGB_COLOR_TYPE, &dropColor, &dataSize)
			== B_OK){
			
		// calculate view relative drop point
		BPoint dropPoint = ConvertFromScreen(message->DropPoint());
		
		// calculate current keypad and display areas
		float sizeDisp = mHeight * K_DISPLAY_YPROP;
		// BRect dispRect(0.0, 0.0, mWidth, sizeDisp);
		BRect keypadRect(0.0, sizeDisp, mWidth, mHeight);
		
		// check location of color drop
		if (keypadRect.Contains(dropPoint)){
		
			// recolor GUI
			if (dropColor){
				mRGBBaseColor = * ((rgb_color *) dropColor);
				} // end if
			Colorize();
			
			// redraw keypad
			Draw(keypadRect);
			
			} // end if

		} // end if

	else {

		// look for text/plain MIME data
		const void *text;
		ssize_t numBytes;
		if (message->FindData("text/plain", B_MIME_TYPE, &text, &numBytes) == B_OK){
			mDisplayText->Append((const char *) text, numBytes);
			} // end if
	
		// redraw display
		float sizeDisp = mHeight * K_DISPLAY_YPROP;
		BRect displayRect(0.0, 0.0, mWidth, sizeDisp);
		Draw(displayRect);
		
		} // end else

} // end CalcView::Paste()


/*******************************************************************************
<func>
Class:		CalcView
Function:	void ParseCalcDesc(const char *calcDesc);
Summary:	Parse calculator description for view.
Params:		calcDesc	calculator description string
Author:		timmy
Created:	1997.10.02
History:	
</func>
*******************************************************************************/
void CalcView::ParseCalcDesc(const char *calcDesc){

	// TODO: should calculate dimensions from desc here!
	mKeypad = new CalcKey[mRows * mCols];

	// scan through calculator description and assemble keypad
	bool scanFlag = true;
	CalcKey *key = mKeypad;
	const char *p = calcDesc;
	while (scanFlag){
	
		// copy label
		char *l = key->label;
		while (!isspace(*p)){
			*l++ = *p++;
			} // end while
		*l = '\0';
		
		// set code
		if (strcmp(key->label, "=") == 0) strcpy(key->code, "\n");
		else strcpy(key->code, key->label);
		
		// set keymap
		if (strlen(key->label)==1){
			strcpy(key->keymap, key->label);
			} // end if
		else {
			*key->keymap = '\0';
			} // end if
		
		// advance	
		while (isspace(*p)) ++p;
		key++;
		
		// check desc termination
		if (*p == '\0') scanFlag = false;
		
		} // end while

} // end CalcView::ParseCalcDesc()


/*******************************************************************************
<func>
Class:		CalcView
Function:	void PressKey(int key);
Summary:	Press a key, any key.
Params:		key		index of key to press
Author:		timmy
Created:	1997.10.05
History:
1997.10.28 timmy Added mOptions check before making sound.
</func>
*******************************************************************************/
void CalcView::PressKey(int key){

	assert(key < (mRows * mCols));
	assert(key >= 0);

	// check for backspace
	if (strcmp(mKeypad[key].label, "BS") == 0){
		if (mDisplayText->Length() > 0){
			mDisplayText->Truncate(mDisplayText->Length() - 1);
			} // end if
		else {
			beep();
			return; // no need to redraw
			} // end else
		} // end if
		
	// check for clear
	else if (strcmp(mKeypad[key].label, "C") == 0){
		mDisplayText->SetTo("");
		} // end else if
		
	else {

		// append to display text
		mDisplayText->Append(mKeypad[key].code);
	
		// check for evaluation order
		if (mKeypad[key].code[0] == '\n'){
			Evaluate();
			} // end if
		
		else {
		
			// audio feedback
			if (mOptions->mAudioFeedback){
				BEntry zimp("key.AIFF");
				entry_ref zimp_ref;
				zimp.GetRef(&zimp_ref);
				play_sound(&zimp_ref, true, false, true);
				} // end if
		
			} // end else

		} // end if

	// redraw display
	float sizeDisp = mHeight * K_DISPLAY_YPROP;
	BRect displayRect(0.0, 0.0, mWidth, sizeDisp);
	Draw(displayRect);

} // end CalcView::PressKey()


/*******************************************************************************
<func>
Class:		CalcView
Function:	void PressKey(char *label);
Summary:	Press a key by label.
Params:		label	text label of the key to press
Author:		timmy
Created:	1997.10.05
History:	-
</func>
*******************************************************************************/
void CalcView::PressKey(char *label){

	int keys = mRows * mCols;
	for (int i=0; i<keys; i++){
		if (strcmp(mKeypad[i].label, label) == 0){
			PressKey(i);
			return;
			} // end if
		} // end for

} // end CalcView::PressKey()


/*******************************************************************************
<func>
Class:		CalcView
Function:	void Colorize();
Summary:	Sort out the color scheme.	
Author:		timmy
Created:	1997.10.12
Notes:		-
History:
1999.05.30 timmy Added clamping. Fixed GCC warning.
</func>
*******************************************************************************/
void CalcView::Colorize(){

	float f;

	// assemble color scheme
	f = ((float) mRGBBaseColor.red) * 1.25f + 0.5f;
	mRGBLightEdge.red = (f > 255.0) ? 255 : (uint8) f;
	f = ((float) mRGBBaseColor.green) * 1.25f + 0.5f;
	mRGBLightEdge.green = (f > 255.0) ? 255 : (uint8) f;
	f = ((float) mRGBBaseColor.blue) * 1.25f + 0.5f;
	mRGBLightEdge.blue = (f > 255.0) ? 255 : (uint8) f;
	mRGBLightEdge.alpha = 255;
	
	mRGBDarkEdge.red = (uint8)((float) mRGBBaseColor.red * 0.75f + 0.5f);
	mRGBDarkEdge.green = (uint8)((float) mRGBBaseColor.green * 0.75f + 0.5f);
	mRGBDarkEdge.blue = (uint8)((float) mRGBBaseColor.blue * 0.75f + 0.5f);
	mRGBDarkEdge.alpha = 255;
	
	mRGBDisplay.red = 0;
	mRGBDisplay.green = 0;
	mRGBDisplay.blue = 0;
	mRGBDisplay.alpha = 255;

} // end CalcView::Colorize()


/*******************************************************************************
<func>
Class:		CalcView
Function:	void Evaluate();
Summary:	Evaluate current expression.
Author:		timmy
Created:	1997.10.15
Notes:
Numerical evaluation is performed by Ian Ollman's excellent Parser class. See
Parser.h/cpp for more info.
History:
1997.10.16 timmy Removed debug printf().
1997.10.28 timmy Added options check before audio feedback.
1997.10.28 timmy Added intelligent result formating. Added empty string check.
1998.11.05 timmy Rewrote to use Parser.h rather than libcalc.
1999.05.30 timmy Fixed ANSI C++ violation. Adjustments for GCC.
</func>
*******************************************************************************/
void CalcView::Evaluate(){

	const float EXP_SWITCH_HI = 1e12; // # digits to switch from std->exp form
	const float EXP_SWITCH_LO = 1e-12;

	if (mDisplayText->Length() == 0){
		beep();
		return;
		} // end if

	// audio feedback
	if (mOptions->mAudioFeedback){
		BEntry zimp("zimp.AIFF");
		entry_ref zimp_ref;
		zimp.GetRef(&zimp_ref);
		play_sound(&zimp_ref, true, false, false);
		} // end if

	FunctionInterpreter		*interpreter = NULL;
	ParseInfo				parseInfo;
	double					value;

	// printf("evaluate: %s\n", mDisplayText->String());
	try {
		char *tmpstr = strdup(mDisplayText->String());
		parseInfo.segmentStart = parseInfo.string = tmpstr;
		parseInfo.segmentEnd =
			parseInfo.segmentStart + mDisplayText->Length() - 1;
		interpreter = ::new FunctionInterpreter(&parseInfo);
		value = interpreter->Evaluate();
		// printf(" = %f\n", value);
		// ::delete interpreter;
		free(tmpstr);
		} // end try

	catch (ParseErrorMsg error){
		//HandleParseError(error);
		puts("Error!\n");
		*mDisplayText = "error!";
		return;
		} // end catch
		
	char buf[64];
	if (value == 0) strcpy(buf, "0");
	else if (((value <  EXP_SWITCH_HI) && (value >  EXP_SWITCH_LO)) ||
		((value > -EXP_SWITCH_HI) && (value < -EXP_SWITCH_LO))){
		sprintf(buf, "%9f", value); // print in std form
		// hack to remove surplus zeros!
		if (strchr(buf, '.')){
			int i; // loop counter
			for (i = strlen(buf) - 1; i > 0; i--){
				if (buf[i] == '0') buf[i] = '\0';
				else break;
				} // end for
			if (buf[i] == '.') buf[i] = '\0';
			} // end if
		} // end if
	else sprintf(buf, "%e", value); // print in exponential form
		
	// render new result to display
	*mDisplayText = buf;
	
	// redraw display
	float sizeDisp = mHeight * K_DISPLAY_YPROP;
	BRect displayRect(0.0, 0.0, mWidth, sizeDisp);
	Draw(displayRect);
	
} // end CalcView::Evaluate()


/*******************************************************************************
<func>
Class:		CalcView
Function:	void CreatePopUpMenu();
Summary:	Create pop-up menu system.
Author:		timmy
Created:	1997.10.14
Notes:		Pop-up menu appears when user right-clicks anywhere in CalcView.
History:
1997.10.18 timmy Moved into CalcView class.
</func>
*******************************************************************************/
void CalcView::CreatePopUpMenu(){

	// construct items
	mAboutItem = new BMenuItem("About Calculator...",
		new BMessage(B_ABOUT_REQUESTED));
	mOptionsItem = new BMenuItem("Options...",
		new BMessage(K_OPTIONS_REQUESTED));

	// construct menu
	mPopUpMenu = new BPopUpMenu("pop-up", false, false);
	mPopUpMenu->AddItem(mAboutItem);
	mPopUpMenu->AddItem(mOptionsItem);

	// add to window ?

} // end CalcView::CreatePopUpMenu()


/******************************************************************************/

// end CalcView.cpp
